
#include "E.h"

#define WIN_LAYER_DESKTOP                0
#define WIN_LAYER_BELOW                  2
#define WIN_LAYER_NORMAL                 4
#define WIN_LAYER_ONTOP                  6
#define WIN_LAYER_DOCK                   8
#define WIN_LAYER_ABOVE_DOCK             10

#define WIN_STATE_STICKY          (1<<0)	/* everyone knows sticky */
#define WIN_STATE_RESERVED_BIT1   (1<<1)	/* removed minimize here */
#define WIN_STATE_MAXIMIZED_VERT  (1<<2)	/* window in maximized V state */
#define WIN_STATE_MAXIMIZED_HORIZ (1<<3)	/* window in maximized H state */
#define WIN_STATE_HIDDEN          (1<<4)	/* not on taskbar but window visible */
#define WIN_STATE_SHADED          (1<<5)	/* shaded (NeXT style) */
#define WIN_STATE_HID_WORKSPACE   (1<<6)	/* not on current desktop */
#define WIN_STATE_HID_TRANSIENT   (1<<7)	/* owner of transient is hidden */
#define WIN_STATE_FIXED_POSITION  (1<<8)	/* window is fixed in position even */
#define WIN_STATE_ARRANGE_IGNORE  (1<<9)	/* ignore for auto arranging */

/* this is temporary */
static ToolTip     *ttip = NULL;

void                ToolTipTimeout(int val, void *data);

void                addmenu(void);

void
ToolTipTimeout(int val, void *data)
{
  int                 x, y, dum;
  Window              win, rt, ch;
  ActionClass        *ac;

  EDBUG(5, "ToolTipTimeout");

  XQueryPointer(disp, root.win, &rt, &ch, &x, &y, &dum,
		&dum, (unsigned int *)&dum);
  win = WindowAtXY(x, y);
  ac = FindActionClass(win);
  if (!ac)
    EDBUG_RETURN_;

  if (!ttip)
    ttip = FindItem("DEFAULT", 0, LIST_FINDBY_NAME,
		    LIST_TYPE_TOOLTIP);
  if (ac->tooltipstring)
    ShowToolTip(ttip, "DO_HELP_BUTTON", ac->tooltipstring, x, y);
  EDBUG_RETURN_;
  data = NULL;
  val = 0;
}

/* end temporary stuff */

void
HandleClientMessage(XEvent * ev)
{
  Atom                a;
  EWin               *ewin;
  int                 d, xo, yo;

  EDBUG(5, "HandleClientMessage");
  a = XInternAtom(disp, "ENL_MSG", False);
  if (ev->xclient.message_type == a)
    {
      HandleComms(ev);
      EDBUG_RETURN_;
    }
  a = XInternAtom(disp, "_WIN_AREA", False);
  if (ev->xclient.message_type == a)
    {
      SetCurentArea(ev->xclient.data.l[0], ev->xclient.data.l[1]);
      EDBUG_RETURN_;
    }
  a = XInternAtom(disp, "_WIN_WORKSPACE", False);
  if (ev->xclient.message_type == a)
    {
      GotoDesktop(ev->xclient.data.l[0]);
      EDBUG_RETURN_;
    }
  a = XInternAtom(disp, "_WIN_LAYER", False);
  if (ev->xclient.message_type == a)
    {
      ewin = FindEwinByChildren(ev->xclient.window);
      if (ewin)
	{
	  ewin->layer = ev->xclient.data.l[0];
	  XChangeProperty(disp, ewin->win, a, XA_CARDINAL, 32,
			  PropModeReplace,
			  (unsigned char *)(&(ev->xclient.data.l[0])), 1);
	  RaiseEwin(ewin);
	}
      EDBUG_RETURN_;
    }
  a = XInternAtom(disp, "_WIN_STATE", False);
  if (ev->xclient.message_type == a)
    {
      ewin = FindEwinByChildren(ev->xclient.window);
      if (!ewin)
	EDBUG_RETURN_;
      if (ev->xclient.data.l[0] & WIN_STATE_FIXED_POSITION)
	{
	  if (ev->xclient.data.l[1] & WIN_STATE_FIXED_POSITION)
	    ewin->fixedpos = 1;
	  else
	    ewin->fixedpos = 0;
	}
      if (ev->xclient.data.l[0] & WIN_STATE_ARRANGE_IGNORE)
	{
	  if (ev->xclient.data.l[1] & WIN_STATE_ARRANGE_IGNORE)
	    ewin->ignorearrange = 1;
	  else
	    ewin->ignorearrange = 0;
	}
      if (ev->xclient.data.l[0] & WIN_STATE_STICKY)
	{
	  if (ev->xclient.data.l[1] & WIN_STATE_STICKY)
	    {
	      if (!(ewin->sticky))
		{
		  xo = desks.desk[ewin->desktop].x;
		  yo = desks.desk[ewin->desktop].y;
		  MoveEwinToDesktopAt(ewin, 0, ewin->x + xo, ewin->y + yo);
		  DesktopRemoveEwin(ewin);
		  ewin->sticky = 1;
		  RaiseEwin(ewin);
		  DrawEwin(ewin);
		  DesktopRemoveEwin(ewin);
		  ApplySclass(FindItem("SOUND_WINDOW_STICK", 0,
				       LIST_FINDBY_NAME, LIST_TYPE_SCLASS));
		}
	    }
	  else
	    {
	      if (ewin->sticky)
		{
		  ewin->sticky = 0;
		  d = DesktopAt(ewin->x, ewin->y);
		  xo = desks.desk[d].x;
		  yo = desks.desk[d].y;
		  MoveEwinToDesktopAt(ewin, d, ewin->x - xo, ewin->y - yo);
		  RaiseEwin(ewin);
		  DrawEwin(ewin);
		  ApplySclass(FindItem("SOUND_WINDOW_UNSTICK", 0,
				       LIST_FINDBY_NAME, LIST_TYPE_SCLASS));
		}
	    }
	}
      if (ev->xclient.data.l[0] & WIN_STATE_SHADED)
	{
	  if (ev->xclient.data.l[1] & WIN_STATE_SHADED)
	    {
	      ShadeEwin(ewin);
	    }
	  else
	    {
	      UnShadeEwin(ewin);
	    }
	}
      GNOME_SetHint(ewin);
      EDBUG_RETURN_;
    }
  a = XInternAtom(disp, "WM_CHANGE_STATE", False);
  if (ev->xclient.message_type == a)
    {
      ewin = FindEwinByChildren(ev->xclient.window);
      if ((ewin) && (ev->xclient.data.l[0] == IconicState) &&
	  (!(ewin->iconified)))
	{
	  HideEwin(ewin);
	  ICCCM_Iconify(ewin);
	  MakeIcon(ewin);
	}
    }
  EDBUG_RETURN_;
}

void
HandleFocusWindowIn(Window win)
{
  EWin               *ewin;

  EDBUG(5, "HandleFocusWindow");
  if (just_flipped)
    {
      just_flipped = 0;
      EDBUG_RETURN_;
    }
  ewin = FindItem(NULL, win, LIST_FINDBY_ID, LIST_TYPE_EWIN);
  if (ewin != mode.focuswin)
    {
      if (mode.focuswin)
	{
	  mode.focuswin->active = 0;
	  DrawEwin(mode.focuswin);
	  if (mode.focusmode == FOCUS_CLICK)
	    XGrabButton(disp, AnyButton, 0, mode.focuswin->win_container, False,
			ButtonPressMask,
			GrabModeSync, GrabModeAsync, None, None);
	}
      mode.focuswin = ewin;
      if (mode.focuswin)
	{
	  mode.focuswin->active = 1;
	  DrawEwin(mode.focuswin);
	  if (mode.focusmode == FOCUS_CLICK)
	    XUngrabButton(disp, AnyButton, 0, mode.focuswin->win_container);
	}
    }
  EDBUG_RETURN_;
}

void
HandleFocusWindow(Window win)
{
  EWin               *found_ewin;
  char                isclient;

  EDBUG(5, "HandleFocusWindow");
  found_ewin = FindEwinByChildren(win);
  isclient = 0;
  if (mode.focusmode == FOCUS_CLICK)
    {
      mode.mouse_over_win = found_ewin;
    }
  else if (mode.focusmode == FOCUS_SLOPPY)
    {
      if (found_ewin != mode.focuswin)
	{
	  if (found_ewin)
	    {
	      FocusToEWin(found_ewin);
	    }
	}
      mode.mouse_over_win = found_ewin;
    }
  else if (mode.focusmode == FOCUS_POINTER)
    {
      if (!found_ewin)
	found_ewin = GetEwinPointerInClient();
      if (found_ewin != mode.focuswin)
	{
	  FocusToEWin(found_ewin);
	}
      mode.mouse_over_win = found_ewin;
    }
  EDBUG_RETURN_;
}

void
HandleChildShapeChange(XEvent * ev)
{
  Window              win;
  EWin               *ewin;
  Border             *b;

  EDBUG(5, "HandleChildShapeChange");
  win = ((XShapeEvent *) ev)->window;
  ewin = FindItem(NULL, win, LIST_FINDBY_ID, LIST_TYPE_EWIN);
  if (ewin)
    {
      b = ewin->border;
      SyncBorderToEwin(ewin);
      if (ewin->border == b)
	PropagateShapes(ewin->win);
    }
  EDBUG_RETURN_;
}

void
HandleMotion(XEvent * ev)
{
  int                 dx, dy;
  int                 pw, ph;
  int                 x, y, w, h;
  EWin               *ewin;

  EDBUG(5, "HandleMotion");

  if (ttip)
    HideToolTip(ttip);
  RemoveTimerEvent("TOOLTIP_TIMEOUT");
  if (mode.tooltips)
    DoIn("TOOLTIP_TIMEOUT", mode.tiptime, ToolTipTimeout, 0, NULL);

  mode.px = mode.x;
  mode.py = mode.y;
  mode.x = ev->xmotion.x_root;
  mode.y = ev->xmotion.y_root;
  ewin = NULL;
  desks.current = DesktopAt(mode.x, mode.y);

  if (!(ev->xmotion.state & (Button1Mask | Button2Mask |
			     Button3Mask | Button4Mask |
			     Button5Mask)))
    {
      switch (mode.mode)
	{
	case MODE_RESIZE:
	case MODE_RESIZE_H:
	case MODE_RESIZE_V:
	  doResizeEnd(NULL);
	  break;
	case MODE_MOVE:
	  doMoveEnd(NULL);
	  break;
	case MODE_BUTTONDRAG:
	  doDragButtonEnd(NULL);
	  break;
	default:
	  break;
	}
      EDBUG_RETURN_;
    }
  switch (mode.mode)
    {
    case MODE_MOVE:
      if (mode.ewin)
	{
	  int                 ndx, ndy;
	  int                 prx, pry;
	  char                dox = 0, doy = 0;

	  ewin = mode.ewin;
	  dx = mode.x - mode.px;
	  dy = mode.y - mode.py;
	  SnapEwin(ewin, dx, dy, &ndx, &ndy);
	  prx = ewin->reqx;
	  pry = ewin->reqy;
	  if ((ndx != dx) &&
	      (((ewin->x == 0) &&
		(!(IN_RANGE(ewin->reqx, ewin->x, mode.screen_snap_dist)))) ||
	       ((ewin->x == (root.w - ewin->w)) &&
		(!(IN_RANGE(ewin->reqx, ewin->x, mode.screen_snap_dist)))) ||
	       ((ewin->x != 0) && (ewin->x != (root.w - ewin->w) &&
		   (!(IN_RANGE(ewin->reqx, ewin->x, mode.edge_snap_dist)))))))
	    {

	      ndx = ewin->reqx - ewin->x;
	      prx = ewin->x;
	      dx = 0;
	      dox = 1;
	    }
	  if ((ndy != dy) &&
	      (((ewin->y == 0) &&
		(!(IN_RANGE(ewin->reqy, ewin->y, mode.screen_snap_dist)))) ||
	       ((ewin->y == (root.h - ewin->h)) &&
		(!(IN_RANGE(ewin->reqy, ewin->y, mode.screen_snap_dist)))) ||
	       ((ewin->y != 0) && (ewin->y != (root.h - ewin->h) &&
		   (!(IN_RANGE(ewin->reqy, ewin->y, mode.edge_snap_dist)))))))
	    {
	      ndy = ewin->reqy - ewin->y;
	      pry = ewin->y;
	      dy = 0;
	      doy = 1;
	    }
	  if (mode.movemode == 0)
	    DetermineEwinFloat(ewin, ndx, ndy);
	  DrawEwinShape(ewin, mode.movemode, ewin->x + ndx, ewin->y + ndy,
			ewin->client.w, ewin->client.h, mode.firstlast);
	  if (!(dox))
	    ewin->reqx = prx + dx;
	  if (!(doy))
	    ewin->reqy = pry + dy;
	}
      break;
    case MODE_RESIZE:
      if (mode.ewin)
	{
	  ewin = mode.ewin;
	  switch (mode.resize_detail)
	    {
	    case 0:
	      pw = ewin->client.w;
	      ph = ewin->client.h;
	      w = mode.win_w - (mode.x - mode.start_x);
	      h = mode.win_h - (mode.y - mode.start_y);
	      x = mode.win_x + (mode.x - mode.start_x);
	      y = mode.win_y + (mode.y - mode.start_y);
	      ewin->client.w = w;
	      ewin->client.h = h;
	      ICCCM_MatchSize(ewin);
	      w = ewin->client.w;
	      h = ewin->client.h;
	      if (pw == ewin->client.w)
		x = ewin->x;
	      else
		x = mode.win_x + mode.win_w - w;
	      if (ph == ewin->client.h)
		y = ewin->y;
	      else
		y = mode.win_y + mode.win_h - h;
	      ewin->client.w = pw;
	      ewin->client.h = ph;
	      DrawEwinShape(ewin, mode.resizemode, x, y, w, h, mode.firstlast);
	      break;
	    case 1:
	      ph = ewin->client.h;
	      w = mode.win_w + (mode.x - mode.start_x);
	      h = mode.win_h - (mode.y - mode.start_y);
	      x = ewin->x;
	      y = mode.win_y + (mode.y - mode.start_y);
	      ewin->client.h = h;
	      ICCCM_MatchSize(ewin);
	      h = ewin->client.h;
	      if (ph == ewin->client.h)
		y = ewin->y;
	      else
		y = mode.win_y + mode.win_h - h;
	      ewin->client.h = ph;
	      DrawEwinShape(ewin, mode.resizemode, x, y, w, h, mode.firstlast);
	      break;
	    case 2:
	      pw = ewin->client.w;
	      w = mode.win_w - (mode.x - mode.start_x);
	      h = mode.win_h + (mode.y - mode.start_y);
	      x = mode.win_x + (mode.x - mode.start_x);
	      y = ewin->y;
	      ewin->client.w = w;
	      ICCCM_MatchSize(ewin);
	      w = ewin->client.w;
	      if (pw == ewin->client.w)
		x = ewin->x;
	      else
		x = mode.win_x + mode.win_w - w;
	      ewin->client.w = pw;
	      DrawEwinShape(ewin, mode.resizemode, x, y, w, h, mode.firstlast);
	      break;
	    case 3:
	      w = mode.win_w + (mode.x - mode.start_x);
	      h = mode.win_h + (mode.y - mode.start_y);
	      x = ewin->x;
	      y = ewin->y;
	      DrawEwinShape(ewin, mode.resizemode, x, y, w, h, mode.firstlast);
	      break;
	    default:
	      break;
	    }
	}
      break;
    case MODE_RESIZE_H:
      if (mode.ewin)
	{
	  ewin = mode.ewin;
	  switch (mode.resize_detail)
	    {
	    case 0:
	      pw = ewin->client.w;
	      w = mode.win_w - (mode.x - mode.start_x);
	      h = ewin->client.h;
	      x = mode.win_x + (mode.x - mode.start_x);
	      y = ewin->y;
	      ewin->client.w = w;
	      ICCCM_MatchSize(ewin);
	      w = ewin->client.w;
	      if (pw == ewin->client.w)
		x = ewin->x;
	      else
		x = mode.win_x + mode.win_w - w;
	      ewin->client.w = pw;
	      DrawEwinShape(ewin, mode.resizemode, x, y, w, h, mode.firstlast);
	      break;
	    case 1:
	      w = mode.win_w + (mode.x - mode.start_x);
	      h = ewin->client.h;
	      x = ewin->x;
	      y = ewin->y;
	      DrawEwinShape(ewin, mode.resizemode, x, y, w, h, mode.firstlast);
	      break;
	    default:
	      break;
	    }
	}
      break;
    case MODE_RESIZE_V:
      if (mode.ewin)
	{
	  ewin = mode.ewin;
	  switch (mode.resize_detail)
	    {
	    case 0:
	      ph = ewin->client.h;
	      w = ewin->client.w;
	      h = mode.win_h - (mode.y - mode.start_y);
	      x = ewin->x;
	      y = mode.win_y + (mode.y - mode.start_y);
	      ewin->client.h = h;
	      ICCCM_MatchSize(ewin);
	      h = ewin->client.h;
	      if (ph == ewin->client.h)
		y = ewin->y;
	      else
		y = mode.win_y + mode.win_h - h;
	      ewin->client.h = ph;
	      DrawEwinShape(ewin, mode.resizemode, x, y, w, h, mode.firstlast);
	      break;
	    case 1:
	      w = ewin->client.w;
	      h = mode.win_h + (mode.y - mode.start_y);
	      x = ewin->x;
	      y = ewin->y;
	      DrawEwinShape(ewin, mode.resizemode, x, y, w, h, mode.firstlast);
	      break;
	    default:
	      break;
	    }
	}
      break;
    case MODE_DESKDRAG:
      dx = mode.x - mode.px;
      dy = mode.y - mode.py;
      switch (desks.dragdir)
	{
	case 0:
	  if ((desks.desk[mode.deskdrag].x + dx) < 0)
	    dx = -desks.desk[mode.deskdrag].x;
	  MoveDesktop(mode.deskdrag, desks.desk[mode.deskdrag].x + dx,
		      desks.desk[mode.deskdrag].y);
	  break;
	case 1:
	  if ((desks.desk[mode.deskdrag].x + dx) > 0)
	    MoveDesktop(mode.deskdrag, 0, desks.desk[mode.deskdrag].y);
	  else
	    MoveDesktop(mode.deskdrag, desks.desk[mode.deskdrag].x + dx,
			desks.desk[mode.deskdrag].y);
	  break;
	case 2:
	  if ((desks.desk[mode.deskdrag].y + dy) < 0)
	    dy = -desks.desk[mode.deskdrag].y;
	  MoveDesktop(mode.deskdrag, desks.desk[mode.deskdrag].x,
		      desks.desk[mode.deskdrag].y + dy);
	  break;
	case 3:
	  if ((desks.desk[mode.deskdrag].y + dy) > 0)
	    MoveDesktop(mode.deskdrag, desks.desk[mode.deskdrag].x, 0);
	  else
	    MoveDesktop(mode.deskdrag, desks.desk[mode.deskdrag].x,
			desks.desk[mode.deskdrag].y + dy);
	  break;
	default:
	  break;
	}
      break;
    case MODE_BUTTONDRAG:
      dx = mode.x - mode.px;
      dy = mode.y - mode.py;
      if (mode.button_move_pending)
	{
	  x = mode.x - mode.start_x;
	  y = mode.y - mode.start_y;
	  if (x < 0)
	    x = -x;
	  if (y < 0)
	    y = -y;
	  if ((x > mode.button_move_resistance) ||
	      (y > mode.button_move_resistance))
	    mode.button_move_pending = 0;
	}
      if (!mode.button_move_pending)
	{
	  if (mode.button)
	    {
	      MovebuttonToCoord(mode.button,
				mode.button->x + dx,
				mode.button->y + dy);
	      if (mode.deskmode == MODE_DESKRAY)
		{
		  MoveDesktop(mode.deskdrag, desks.desk[mode.deskdrag].x,
			      desks.desk[mode.deskdrag].y + dy);
		}
	    }
	}
      break;
    default:
      break;
    }
#define SCROLL_DIST 13
  if ((mode.cur_menu_mode) && (mode.cur_menu_depth > 0))
    {
      int                 i, offx, offy, xdist = 0, ydist = 0;
      EWin               *ewin;

      if (mode.x > (root.w - (SCROLL_DIST + 1)))
	xdist = -(SCROLL_DIST - 1 + (mode.x - root.w));
      else if (mode.x < SCROLL_DIST)
	xdist = SCROLL_DIST - 1 - (mode.x);

      if (mode.y > (root.h - (SCROLL_DIST + 1)))
	ydist = -(SCROLL_DIST - 1 + (mode.y - root.h));
      else if (mode.y < SCROLL_DIST)
	ydist = SCROLL_DIST - 1 - (mode.y);

      if (mode.cur_menu[mode.cur_menu_depth - 1])
	{
	  ewin = FindEwinByMenu(mode.cur_menu[mode.cur_menu_depth - 1]);
	  if (ewin)
	    {
	      if (xdist <= 0)
		offx = root.w - (ewin->x + ewin->w);
	      else
		offx = -ewin->x;
	      if (ydist <= 0)
		offy = root.h - (ewin->y + ewin->h);
	      else
		offy = -ewin->y;
	      if ((xdist <= 0) && (offx <= 0) && (xdist < offx))
		xdist = offx;
	      if ((xdist >= 0) && (offx >= 0) && (xdist > offx))
		xdist = offx;
	      if ((ydist <= 0) && (offy <= 0) && (ydist < offy))
		ydist = offy;
	      if ((ydist >= 0) && (offy >= 0) && (ydist > offy))
		ydist = offy;
	      for (i = 0; i < mode.cur_menu_depth; i++)
		{
		  if (mode.cur_menu[i])
		    {
		      ewin = FindEwinByMenu(mode.cur_menu[i]);
		      if (ewin)
			MoveEwin(ewin, ewin->x + xdist, ewin->y + ydist);
		    }
		}
	    }
	}
      XWarpPointer(disp, None, None, 0, 0, 0, 0, xdist, ydist);
    }
  EDBUG_RETURN_;
}

void
HandleDestroy(XEvent * ev)
{
  Window              win;
  EWin               *ewin;
  Client             *c;

  EDBUG(5, "HandleDestroy");
  win = ev->xdestroywindow.window;
  ewin = RemoveItem(NULL, win, LIST_FINDBY_ID, LIST_TYPE_EWIN);
  if (ewin)
    if (ewin->iconified > 0)
      RemoveMiniIcon(ewin);

  mode.context_win = win;

  if (ewin)
    {
      if (ewin == mode.ewin)
	{
	  if (mode.slideout)
	    HideSlideout(mode.slideout, mode.context_win);
	  switch (mode.mode)
	    {
	    case MODE_RESIZE:
	    case MODE_RESIZE_H:
	    case MODE_RESIZE_V:
	      doResizeEnd(NULL);
	      break;
	    case MODE_MOVE:
	      doMoveEnd(NULL);
	      break;
	    default:
	      break;
	    }
	}
      if (mode.doingslide)
	{
	  DrawEwinShape(ewin, mode.slidemode, ewin->x, ewin->y, ewin->client.w, ewin->client.h, 2);
	  mode.doingslide = 0;
	}
      if (ewin == mode.focuswin)
	FocusToEWin(NULL);
      if (ewin == mode.mouse_over_win)
	mode.mouse_over_win = NULL;
      if (ewin == mode.ewin)
	mode.ewin = NULL;
      if (ewin->docked)
	DockDestroy(ewin);
      DesktopRemoveEwin(ewin);
      FreeEwin(ewin);
      GNOME_SetClientList();
      EDBUG_RETURN_;
    }
  c = FindItem(NULL, win, LIST_FINDBY_ID, LIST_TYPE_CLIENT);
  if (c)
    DeleteClient(c);
  EDBUG_RETURN_;
}

void
HandleProperty(XEvent * ev)
{
  Window              win;
  EWin               *ewin;
  char                title[10240];
  int                 desktop;

  EDBUG(5, "HandleProperty");
  win = ev->xproperty.window;
  ewin = FindItem(NULL, win, LIST_FINDBY_ID, LIST_TYPE_EWIN);
  if (ewin)
    {
      GrabX();
      if (ewin->client.title)
	strncpy(title, ewin->client.title, 10240);
      desktop = ewin->desktop;
      GNOME_GetHintDesktop(ewin);
      if ((desktop != ewin->desktop) && (!ewin->sticky))
	MoveEwinToDesktop(ewin, ewin->desktop);
      GNOME_GetHintIcons(ewin);
      GNOME_GetHintAppState(ewin);
      GNOME_GetExpandedSize(ewin);
      GNOME_GetHint(ewin);
      ICCCM_GetTitle(ewin);
      ICCCM_GetHints(ewin);
      ICCCM_GetInfo(ewin);
      ICCCM_GetColormap(ewin);
      ICCCM_GetGeoms(ewin);
      SessionGetInfo(ewin);
      SyncBorderToEwin(ewin);
      if (ewin->client.title)
	if (strncmp(title, ewin->client.title, 10240))
	  {
	    UpdateBorderInfo(ewin);
	    CalcEwinSizes(ewin);
	  }

      UngrabX();
    }
  EDBUG_RETURN_;
}

void
HandleCirculate(XEvent * ev)
{
  Window              win;
  EWin               *ewin;

  EDBUG(5, "HandleCirculate");
  win = ev->xcirculaterequest.window;
  ewin = FindItem(NULL, win, LIST_FINDBY_ID, LIST_TYPE_EWIN);
  if (ewin)
    {
      if (ev->xcirculaterequest.place == PlaceOnTop)
	RaiseEwin(ewin);
      else
	LowerEwin(ewin);
    }
  else
    {
      if (ev->xcirculaterequest.place == PlaceOnTop)
	XRaiseWindow(disp, win);
      else
	XLowerWindow(disp, win);
    }
  EDBUG_RETURN_;
}

void
HandleReparent(XEvent * ev)
{
  Window              par;
  EWin               *ewin, *ewin2;
  void              **lst;
  int                 i, num, found = 0;

  EDBUG(5, "HandleReparent");
  par = ev->xreparent.window;
  EDBUG_RETURN_;
  ewin = FindItem(NULL, ev->xreparent.window, LIST_FINDBY_ID, LIST_TYPE_EWIN);
  if (!ewin)
    EDBUG_RETURN_;
  lst = ListItemType(&num, LIST_TYPE_EWIN);
  if ((lst) && (num > 0))
    {
      for (i = 0; i < num; i++)
	{
	  ewin2 = (EWin *) lst[i];
	  if (ewin2->win == ev->xreparent.parent)
	    found = 1;
	}
    }
  if (lst)
    Efree(lst);
  if (!found)
    {
      ewin = RemoveItem(NULL, ev->xreparent.window, LIST_FINDBY_ID, LIST_TYPE_EWIN);
      if (ewin)
	{
	  if (ewin == mode.ewin)
	    {
	      switch (mode.mode)
		{
		case MODE_RESIZE:
		case MODE_RESIZE_H:
		case MODE_RESIZE_V:
		  doResizeEnd(NULL);
		  break;
		case MODE_MOVE:
		  doMoveEnd(NULL);
		  break;
		default:
		  break;
		}
	    }
	  ewin->client.win = 0;
	  if ((mode.slideout) && (ewin == mode.ewin))
	    HideSlideout(mode.slideout, mode.context_win);
	  FreeEwin(ewin);
	  GNOME_SetClientList();
	}
    }
  EDBUG_RETURN_;
}

void
HandleConfigureRequest(XEvent * ev)
{
  Window              win, winrel;
  EWin               *ewin, *ewin2;
  int                 x = 0, y = 0, w = 0, h = 0;
  int                 move = 0, resize = 0;
  XWindowChanges      xwc;

  EDBUG(5, "HandleConfigureRequest");
  win = ev->xconfigurerequest.window;
  ewin = FindItem(NULL, win, LIST_FINDBY_ID, LIST_TYPE_EWIN);
  if (ewin)
    {
      x = ewin->x;
      y = ewin->y;
      w = ewin->client.w;
      h = ewin->client.h;
      winrel = 0;
      if (ev->xconfigurerequest.value_mask & CWX)
	{
	  x = ev->xconfigurerequest.x;
	  move = 1;
	}
      if (ev->xconfigurerequest.value_mask & CWY)
	{
	  y = ev->xconfigurerequest.y;
	  move = 1;
	}
      if (ev->xconfigurerequest.value_mask & CWWidth)
	{
	  w = ev->xconfigurerequest.width;
	  resize = 1;
	}
      if (ev->xconfigurerequest.value_mask & CWHeight)
	{
	  h = ev->xconfigurerequest.height;
	  resize = 1;
	}
      if (ev->xconfigurerequest.value_mask & CWSibling)
	{
	  winrel = ev->xconfigurerequest.above;
	}
      if (ev->xconfigurerequest.value_mask & CWStackMode)
	{
	  ewin2 = FindItem(NULL, winrel, LIST_FINDBY_ID, LIST_TYPE_EWIN);
	  if (ewin2)
	    winrel = ewin2->win;
	  xwc.sibling = winrel;
	  xwc.stack_mode = ev->xconfigurerequest.detail;
	  if (xwc.stack_mode == Above)
	    RaiseEwin(ewin);
	  else if (xwc.stack_mode == Below)
	    LowerEwin(ewin);
	  else
	    XConfigureWindow(disp, ewin->win,
			     ev->xconfigurerequest.value_mask &
			     (CWSibling | CWStackMode), &xwc);
	}
      if ((move) && (resize))
	MoveResizeEwin(ewin,
		       x - ewin->border->border.left,
		       y - ewin->border->border.top,
		       w, h);
      else if (move)
	{
	  MoveEwin(ewin,
		   x - ewin->border->border.left,
		   y - ewin->border->border.top);
	  ICCCM_Configure(ewin);
	}
      else if (resize)
	ResizeEwin(ewin, w, h);
    }
  else
    {
      xwc.x = ev->xconfigurerequest.x;
      xwc.y = ev->xconfigurerequest.y;
      xwc.width = ev->xconfigurerequest.width;
      xwc.height = ev->xconfigurerequest.height;
      xwc.border_width = ev->xconfigurerequest.border_width;
      xwc.sibling = ev->xconfigurerequest.above;
      xwc.stack_mode = ev->xconfigurerequest.detail;
      XConfigureWindow(disp, win, ev->xconfigurerequest.value_mask, &xwc);
    }
  EDBUG_RETURN_;
}

void
HandleResizeRequest(XEvent * ev)
{
  Window              win;
  EWin               *ewin;
  int                 w, h;

  EDBUG(5, "HandleResizeRequest");
  win = ev->xresizerequest.window;
  ewin = FindItem(NULL, win, LIST_FINDBY_ID, LIST_TYPE_EWIN);
  if (ewin)
    {
      w = ev->xresizerequest.width;
      h = ev->xresizerequest.height;
      ResizeEwin(ewin, w, h);
    }
  else
    XResizeWindow(disp, win, ev->xresizerequest.width, ev->xresizerequest.height);
  EDBUG_RETURN_;
}

void
HandleUnmap(XEvent * ev)
{
  Window              win;
  EWin               *ewin;

  EDBUG(5, "HandleUnmap");
  win = ev->xunmap.window;
  ewin = FindItem(NULL, win, LIST_FINDBY_ID, LIST_TYPE_EWIN);
  if (ewin)
    {
      if (ewin->iconified > 1)
	{
	  ewin->iconified = ewin->iconified - 1;
	  EDBUG_RETURN_;
	}
    }
  if (ewin)
    {
      if (ewin->docked)
	DockDestroy(ewin);
      if (ewin == mode.ewin)
	{
	  switch (mode.mode)
	    {
	    case MODE_RESIZE:
	    case MODE_RESIZE_H:
	    case MODE_RESIZE_V:
	      doResizeEnd(NULL);
	      break;
	    case MODE_MOVE:
	      doMoveEnd(NULL);
	      break;
	    default:
	      break;
	    }
	}
      if (!ewin->iconified)
	{
	  if ((mode.slideout) && (ewin == mode.ewin))
	    HideSlideout(mode.slideout, mode.context_win);
	}
      if (ewin == mode.focuswin)
	FocusToEWin(NULL);
      if (ewin == mode.mouse_over_win)
	mode.mouse_over_win = NULL;
      if (ewin == mode.ewin)
	mode.ewin = NULL;
      if (!ewin->iconified)
	{
	  XTranslateCoordinates(disp, ewin->client.win, root.win,
				-ewin->border->border.left,
				-ewin->border->border.top,
				&ewin->client.x, &ewin->client.y, &win);
	  XReparentWindow(disp, ewin->client.win, root.win,
			  ewin->client.x, ewin->client.y);
	  XRemoveFromSaveSet(disp, ewin->client.win);
	  RemoveItem(NULL, ewin->client.win, LIST_FINDBY_ID, LIST_TYPE_EWIN);
	  DesktopRemoveEwin(ewin);
	  FreeEwin(ewin);
	  GNOME_SetClientList();
	}
      else
	HideEwin(ewin);
    }
  EDBUG_RETURN_;
}

void
HandleMapRequest(XEvent * ev)
{
  EDBUG(5, "HandleMapRequest");
  AddToFamily(ev->xmap.window);
  GNOME_SetClientList();
  EDBUG_RETURN_;
}

void
HandleExpose(XEvent * ev)
{
  Window              win;
  EWin              **ewin;
  int                 i, j, num;

  EDBUG(5, "HandleExpose");
  if (ev->xexpose.count > 0)
    EDBUG_RETURN_;
  win = ev->xexpose.window;

  ewin = (EWin **) ListItemType(&num, LIST_TYPE_EWIN);
  for (i = 0; i < num; i++)
    {
      for (j = 0; j < ewin[i]->border->num_winparts; j++)
	{
	  if (win == ewin[i]->bits[j].win)
	    {
	      ewin[i]->bits[j].expose = 1;
	      DrawEwinWinpart(ewin[i], j);
	      Efree(ewin);
	      EDBUG_RETURN_;
	    }
	}

    }
  if (ewin)
    Efree(ewin);
  EDBUG_RETURN_;
}

static char         sentpress = 0;
static Window       click_was_in = 0;

void
HandleMouseDown(XEvent * ev)
{
  Window              win;
  EWin               *ewin;
  EWin              **ewins;
  int                 i, j, num;
  Button            **buttons;
  ActionClass        *ac;
  Menu               *m;
  MenuItem           *mi;
  XEvent              ev2;
  int                 desk_click = -1;
  char                double_click = 0;
  static Time         last_time = 0;
  float               mode_double_click_time = 0.25;

  EDBUG(5, "HandleMouseDown");

  if (mode.cur_menu_mode)
    EDBUG_RETURN_;
  if (ttip)
    HideToolTip(ttip);
  RemoveTimerEvent("TOOLTIP_TIMEOUT");
  if (mode.tooltips)
    DoIn("TOOLTIP_TIMEOUT", mode.tiptime, ToolTipTimeout, 0, NULL);

  if (((float)(ev->xbutton.time - last_time) / 1000) < mode_double_click_time)
    double_click = 1;
  last_time = ev->xbutton.time;

  click_was_in = win = ev->xbutton.window;

  mode.x = ev->xbutton.x_root;
  mode.y = ev->xbutton.y_root;

  if (sentpress)
    {
      win = WindowAtXY(ev->xbutton.x_root, ev->xbutton.y_root);
      sentpress = 0;
    }
  mode.context_win = win;
  for (i = 0; i < mode.numdesktops; i++)
    {
      if (win == desks.desk[i].win)
	{
	  desk_click = i;
	  break;
	}
    }
  if (desk_click >= 0)
    {
      XUngrabPointer(disp, CurrentTime);
      if (ev->xbutton.button == 2)
	{
	  mode.cur_menu_mode = 1;
	  addmenu();
	}
      else
	{
	  sentpress = 1;
	  XSendEvent(disp, bpress_win, False, SubstructureNotifyMask, ev);
	}
      EDBUG_RETURN_;
    }
  if (mode.focusmode == FOCUS_CLICK)
    {
      ewin = FindEwinByChildren(win);
/*      
 * Window              rt, ch;
 * int                 dum;
 * int                 d;
 * * 
 * d = DESKTOPS_WRAP_NUM(DesktopAt(mode.x, mode.y));
 * XQueryPointer(disp, desks.desk[d].win, &rt, &ch, &dum, &dum, &dum,
 * &dum, (unsigned int *)&dum);
 * ewin = FindEwinByBase(ch);
 * if (mode.click_focus_grabbed)
 * {
 * XEvent              xev;
 * 
 * XTestFakeButtonEvent(disp, ev->xbutton.button, True, CurrentTime);
 * UnGrabTheButtons();
 * while (XCheckMaskEvent(disp, ButtonReleaseMask, &xev))
 * XTestFakeButtonEvent(disp, xev.xbutton.button, False,
 * CurrentTime);
 * } */
      XUngrabPointer(disp, CurrentTime);
      FocusToEWin(ewin);
      if (ewin)
	RaiseEwin(ewin);
      if ((ewin) && (win == ewin->win_container))
	{
	  while (XCheckTypedWindowEvent(disp, win, ev->type, &ev2));
/*        XEvent              xev; */
/*           UnGrabTheButtons(); */
/*           XTestFakeButtonEvent(disp, ev->xbutton.button, True, CurrentTime); */
/*        XTestFakeButtonEvent(disp, ev->xbutton.button, False, CurrentTime); */
/*        
 * while (XCheckMaskEvent(disp, ButtonReleaseMask, &xev))
 * XTestFakeButtonEvent(disp, xev.xbutton.button, False,
 * CurrentTime); */
	}
    }
  m = FindMenuItem(win, &mi);
  if (m)
    {
      mode.cur_menu_mode = 1;
      mi->state = STATE_CLICKED;
      DrawMenuItem(m, mi, 1);
      if (mi->child)
	{
	  int                 mx, my;
	  unsigned int        mw, mh;
	  EWin               *ewin2;

	  mode.cur_menu[0] = m;
	  mode.cur_menu_depth = 1;
	  ShowMenuMasker(m);
	  XUngrabPointer(disp, CurrentTime);
	  ewin = FindEwinByMenu(m);
	  if (ewin)
	    {
	      GetWinXY(mi->win, &mx, &my);
	      GetWinWH(mi->win, &mw, &mh);
	      ShowMenu(mi->child, 1);
	      ewin2 = FindEwinByMenu(mi->child);
	      if (ewin2)
		{
		  MoveEwin(ewin2, ewin->x + ewin->border->border.left + mx + mw,
			   ewin->y + ewin->border->border.top + my -
			   ewin2->border->border.top);
/*                
 * if (ewin2->x + ewin2->w > root.w)
 * MoveEwin(ewin2, root.w - ewin2->w, ewin2->y);
 * if (ewin2->y + ewin2->h > root.h)
 * MoveEwin(ewin2, ewin2->x, root.h - ewin2->h);
 * if (ewin->x < 0)
 * MoveEwin(ewin2, 0, ewin2->y);
 * if (ewin->y < 0)
 * MoveEwin(ewin2, ewin2->x, 0);
 */
		  RaiseEwin(ewin2);
		  ShowEwin(ewin2);
		  if (mode.menuslide)
		    UnShadeEwin(ewin2);
		  mode.cur_menu[mode.cur_menu_depth++] = mi->child;
		}
	    }
	}
      EDBUG_RETURN_;
    }
  if (double_click)
    ev->xbutton.time = 0;

  ewins = (EWin **) ListItemType(&num, LIST_TYPE_EWIN);
  if (ewins)
    {
      for (i = 0; i < num; i++)
	{
	  for (j = 0; j < ewins[i]->border->num_winparts; j++)
	    {
	      if (win == ewins[i]->bits[j].win)
		{
		  mode.ewin = ewins[i];
		  ewins[i]->bits[j].state = STATE_CLICKED;
		  ChangeEwinWinpart(ewins[i], j);
		  if (ewins[i]->border->part[j].aclass)
		    EventAclass(ev, ewins[i]->border->part[j].aclass);
		  Efree(ewins);
		  EDBUG_RETURN_;
		}
	    }
	}
      Efree(ewins);
    }
  buttons = (Button **) ListItemType(&num, LIST_TYPE_BUTTON);
  if (buttons)
    {
      for (i = 0; i < num; i++)
	{
	  if ((win == buttons[i]->win) || (win == buttons[i]->event_win))
	    {
	      if (buttons[i]->inside_win)
		XSendEvent(disp, buttons[i]->inside_win, False, ButtonPressMask, ev);
	      mode.button = buttons[i];
	      buttons[i]->state = STATE_CLICKED;
	      DrawButton(buttons[i]);
	      ac = FindItem("ACTION_BUTTON_DRAG", 0, LIST_FINDBY_NAME,
			    LIST_TYPE_ACLASS);
	      if (ac)
		EventAclass(ev, ac);
	      if (buttons[i]->aclass)
		EventAclass(ev, buttons[i]->aclass);
	      Efree(buttons);
	      EDBUG_RETURN_;
	    }
	}
      Efree(buttons);
    }
  EDBUG_RETURN_;
}

void
HandleMouseUp(XEvent * ev)
{
  Window              win, win2;
  EWin               *ewin;
  EWin              **ewins;
  int                 i, j, num;
  Button            **buttons;
  char                wasdrag;
  char                wasmovres;
  Slideout           *pslideout;
  Menu               *m;
  MenuItem           *mi;

  EDBUG(5, "HandleMouseUp");

  if (ttip)
    HideToolTip(ttip);
  RemoveTimerEvent("TOOLTIP_TIMEOUT");
  if (mode.tooltips)
    DoIn("TOOLTIP_TIMEOUT", mode.tiptime, ToolTipTimeout, 0, NULL);

  win2 = WindowAtXY(ev->xbutton.x_root, ev->xbutton.y_root);
  win = ev->xbutton.window;

  wasdrag = 0;
  wasmovres = 0;

  switch (mode.mode)
    {
    case MODE_RESIZE:
    case MODE_RESIZE_H:
    case MODE_RESIZE_V:
      doResizeEnd(NULL);
      wasmovres = 1;
      break;
    case MODE_MOVE:
      ewin = GetEwin();
      if ((mode.movemode == 0) && (ewin) && (mode.mode == MODE_MOVE))
	DetermineEwinFloat(ewin, 0, 0);
      doMoveEnd(NULL);
      wasmovres = 1;
      break;
    default:
      break;
    }

  if (sentpress)
    {
      sentpress = 0;
      XSendEvent(disp, bpress_win, False, SubstructureNotifyMask, ev);
    }
  mode.context_win = click_was_in;
  mode.destroy = 0;
  pslideout = mode.slideout;
  if (mode.slideout)
    {
      ewin = FindEwinByChildren(mode.slideout->from_win);
      if (ewin)
	mode.ewin = ewin;
    }
  if (mode.mode == MODE_DESKDRAG)
    mode.mode = MODE_NONE;
  if (mode.mode == MODE_BUTTONDRAG)
    {
      if (!mode.button_move_pending)
	wasdrag = 1;
      doDragButtonEnd(NULL);
    }
  m = FindMenuItem(win, &mi);
  if (m)
    {
      mi->state = STATE_HILITED;
      DrawMenuItem(m, mi, 1);
      if (mi->aclass)
	{
	  for (i = 0; i < mi->aclass->num; i++)
	    {
	      Action             *act;

	      act = mi->aclass->list[i];
	      handleAction(act->action);
	    }
	}
    }
  if (mode.cur_menu_mode)
    {
      if (!m)
	{
	  Window              ww;

	  ww = WindowAtXY(mode.x, mode.y);
	  if ((ewin = FindEwinByChildren(ww)))
	    {
	      if (ewin->menu)
		ewin->menu->stuck = 1;
	    }
	}
      for (i = 0; i < mode.cur_menu_depth; i++)
	{
	  if (!mode.cur_menu[i]->stuck)
	    HideMenu(mode.cur_menu[i]);
	}
      HideMenuMasker();
      mode.cur_menu_depth = 0;
      mode.cur_menu_mode = 0;
    }
  ewins = (EWin **) ListItemType(&num, LIST_TYPE_EWIN);
  for (i = 0; i < num; i++)
    {
      for (j = 0; j < ewins[i]->border->num_winparts; j++)
	{
	  if (click_was_in == ewins[i]->bits[j].win)
	    {
	      if (ewins[i] == mode.ewin)
		{
		  switch (mode.mode)
		    {
		    case MODE_RESIZE:
		    case MODE_RESIZE_H:
		    case MODE_RESIZE_V:
		      doResizeEnd(NULL);
		      break;
		    case MODE_MOVE:
		      doMoveEnd(NULL);
		      break;
		    default:
		      break;
		    }
		}
	      if ((ewins[i]->bits[j].state == STATE_CLICKED) &&
		  (!ewins[i]->bits[j].left))
		ewins[i]->bits[j].state = STATE_HILITED;
	      else
		ewins[i]->bits[j].state = STATE_NORMAL;
	      ewins[i]->bits[j].left = 0;
	      ChangeEwinWinpart(ewins[i], j);
	      if ((click_was_in == win2) && (ewins[i]->border->part[j].aclass) &&
		  (!wasmovres))
		EventAclass(ev, ewins[i]->border->part[j].aclass);
	      if ((mode.slideout) && (pslideout))
		HideSlideout(mode.slideout, mode.context_win);
	      Efree(ewins);
	      click_was_in = 0;
	      EDBUG_RETURN_;
	    }
	}
    }
  if (ewins)
    Efree(ewins);
  buttons = (Button **) ListItemType(&num, LIST_TYPE_BUTTON);
  for (i = 0; i < num; i++)
    {
      if ((click_was_in == buttons[i]->win) || (click_was_in == buttons[i]->event_win))
	{
	  if ((buttons[i]->inside_win) && (!wasmovres))
	    XSendEvent(disp, buttons[i]->inside_win, False,
		       ButtonReleaseMask, ev);
	  mode.button = buttons[i];
	  if ((buttons[i]->state == STATE_CLICKED) &&
	      (!buttons[i]->left))
	    buttons[i]->state = STATE_HILITED;
	  else
	    buttons[i]->state = STATE_NORMAL;
	  buttons[i]->left = 0;
	  DrawButton(buttons[i]);
	  if ((click_was_in == win2) && (buttons[i]->aclass) && (!wasdrag) &&
	      (!wasmovres))
	    EventAclass(ev, buttons[i]->aclass);
	  mode.destroy = 0;
	  if ((mode.slideout) && (pslideout))
	    HideSlideout(mode.slideout, mode.context_win);
	  Efree(buttons);
	  click_was_in = 0;
	  EDBUG_RETURN_;
	}
    }
  if (buttons)
    Efree(buttons);
  mode.destroy = 0;
  if ((mode.slideout) && (pslideout))
    HideSlideout(mode.slideout, mode.context_win);
  click_was_in = 0;
  EDBUG_RETURN_;
}

void
HandleMouseIn(XEvent * ev)
{
  Window              win;
  EWin               *ewin;
  EWin              **ewins;
  int                 i, j, num;
  Button            **buttons;
  Menu               *m;
  MenuItem           *mi;

  EDBUG(5, "HandleMouseIn");

  if (ttip)
    HideToolTip(ttip);
  RemoveTimerEvent("TOOLTIP_TIMEOUT");
  if (mode.tooltips)
    DoIn("TOOLTIP_TIMEOUT", mode.tiptime, ToolTipTimeout, 0, NULL);

  win = ev->xcrossing.window;

  mode.context_win = win;

  m = FindMenuItem(win, &mi);
  if (m)
    {
      int                 j;

      if ((win == mi->icon_win) && (ev->xcrossing.detail == NotifyAncestor))
	EDBUG_RETURN_;
      if ((win == mi->win) && (ev->xcrossing.detail == NotifyInferior))
	EDBUG_RETURN_;
      mi->state = STATE_HILITED;
      DrawMenuItem(m, mi, 1);

      for (i = 0; i < mode.cur_menu_depth; i++)
	{
	  if (mode.cur_menu[i] == m)
	    {
	      if ((!mi->child) ||
		  ((mi->child) && (mode.cur_menu[i + 1] != mi->child)))
		{
		  for (j = i + 1; j < mode.cur_menu_depth; j++)
		    HideMenu(mode.cur_menu[j]);
		  mode.cur_menu_depth = i + 1;
		  i = mode.cur_menu_depth;
		  break;
		}
	    }
	}
      if ((mi->child) && (mode.cur_menu_mode))
	{
	  int                 mx, my;
	  unsigned int        mw, mh;
	  EWin               *ewin2;

	  ewin = FindEwinByMenu(m);
	  if (ewin)
	    {
	      GetWinXY(mi->win, &mx, &my);
	      GetWinWH(mi->win, &mw, &mh);
	      ShowMenu(mi->child, 1);
	      ewin2 = FindEwinByMenu(mi->child);
	      if (ewin2)
		{
		  MoveEwin(ewin2, ewin->x + ewin->border->border.left + mx + mw,
			   ewin->y + ewin->border->border.top + my -
			   ewin2->border->border.top);
/*                
 * if (ewin2->x + ewin2->w > root.w)
 * MoveEwin(ewin2, root.w - ewin2->w, ewin2->y);
 * if (ewin2->y + ewin2->h > root.h)
 * MoveEwin(ewin2, ewin2->x, root.h - ewin2->h);
 * if (ewin->x < 0)
 * MoveEwin(ewin2, 0, ewin2->y);
 * if (ewin->y < 0)
 * MoveEwin(ewin2, ewin2->x, 0);
 */
		  RaiseEwin(ewin2);
		  ShowEwin(ewin2);
		  if (mode.menuslide)
		    UnShadeEwin(ewin2);
		  if (mode.cur_menu[mode.cur_menu_depth - 1] != mi->child)
		    mode.cur_menu[mode.cur_menu_depth++] = mi->child;
		}
	    }
	}
      EDBUG_RETURN_;
    }
  ewins = (EWin **) ListItemType(&num, LIST_TYPE_EWIN);
  for (i = 0; i < num; i++)
    {
      for (j = 0; j < ewins[i]->border->num_winparts; j++)
	{
	  if (win == ewins[i]->bits[j].win)
	    {
	      mode.noewin = 0;
	      mode.ewin = ewins[i];
	      if (ewins[i]->bits[j].state == STATE_CLICKED)
		ewins[i]->bits[j].left = 0;
	      else
		{
		  ewins[i]->bits[j].state = STATE_HILITED;
		  ChangeEwinWinpart(ewins[i], j);
		  if (ewins[i]->border->part[j].aclass)
		    EventAclass(ev, ewins[i]->border->part[j].aclass);
		}
	      Efree(ewins);
	      EDBUG_RETURN_;
	    }
	}
    }
  if (ewins)
    Efree(ewins);
  buttons = (Button **) ListItemType(&num, LIST_TYPE_BUTTON);
  for (i = 0; i < num; i++)
    {
      if ((win == buttons[i]->win) || (win == buttons[i]->event_win))
	{
/*               if (b->inside_win)
 * XSendEvent(disp, b->inside_win, False, EnterWindowMask, ev); */
	  mode.button = buttons[i];
	  if (buttons[i]->state == STATE_CLICKED)
	    buttons[i]->left = 0;
	  else
	    {
	      buttons[i]->state = STATE_HILITED;
	      DrawButton(buttons[i]);
	      if (buttons[i]->aclass)
		EventAclass(ev, buttons[i]->aclass);
	    }
	  Efree(buttons);
	  EDBUG_RETURN_;
	}
    }
  if (buttons)
    Efree(buttons);
  EDBUG_RETURN_;
}

void
HandleMouseOut(XEvent * ev)
{
  Window              win;
  EWin              **ewins;
  int                 i, j, num;
  Button            **buttons;
  Menu               *m;
  MenuItem           *mi;

  EDBUG(5, "HandleMouseOut");

  if (ttip)
    HideToolTip(ttip);
  RemoveTimerEvent("TOOLTIP_TIMEOUT");
  if (mode.tooltips)
    DoIn("TOOLTIP_TIMEOUT", mode.tiptime, ToolTipTimeout, 0, NULL);

  win = ev->xcrossing.window;

  mode.context_win = win;

  m = FindMenuItem(win, &mi);
  if (m)
    {
      if ((win == mi->icon_win) && (ev->xcrossing.detail == NotifyAncestor))
	EDBUG_RETURN_;
      if ((win == mi->win) && (ev->xcrossing.detail == NotifyInferior))
	EDBUG_RETURN_;
      mi->state = STATE_NORMAL;
      DrawMenuItem(m, mi, 1);
      EDBUG_RETURN_;
    }
  ewins = (EWin **) ListItemType(&num, LIST_TYPE_EWIN);
  for (i = 0; i < num; i++)
    {
      for (j = 0; j < ewins[i]->border->num_winparts; j++)
	{
	  if (win == ewins[i]->bits[j].win)
	    {
	      if (mode.mode == MODE_NONE)
		mode.ewin = NULL;
	      else
		mode.noewin = 1;
	      if (ewins[i]->bits[j].state == STATE_CLICKED)
		ewins[i]->bits[j].left = 1;
	      else
		{
		  ewins[i]->bits[j].state = STATE_NORMAL;
		  ChangeEwinWinpart(ewins[i], j);
		  if (ewins[i]->border->part[j].aclass)
		    EventAclass(ev, ewins[i]->border->part[j].aclass);
		}
	      Efree(ewins);
	      EDBUG_RETURN_;
	    }
	}
    }
  if (ewins)
    Efree(ewins);
  buttons = (Button **) ListItemType(&num, LIST_TYPE_BUTTON);
  for (i = 0; i < num; i++)
    {
      if ((win == buttons[i]->win) || (win == buttons[i]->event_win))
	{
/*               if (b->inside_win)
 * XSendEvent(disp, b->inside_win, False, LeaveWindowMask, ev); */
	  mode.button = NULL;
	  if (buttons[i]->state == STATE_CLICKED)
	    buttons[i]->left = 1;
	  else
	    {
	      buttons[i]->state = STATE_NORMAL;
	      DrawButton(buttons[i]);
	      if (buttons[i]->aclass)
		EventAclass(ev, buttons[i]->aclass);
	    }
	  Efree(buttons);
	  EDBUG_RETURN_;
	}
    }
  if (buttons)
    Efree(buttons);
  EDBUG_RETURN_;
}
